import 'dart:math';

import 'package:flutter/cupertino.dart';
import 'package:flutter/rendering.dart';
import 'package:nueli/nueli.dart';

class Scroll2 extends Struct {
  final Axis axis;
  final Scroll2Controller controller;
  final Widget child;
  final ScrollPhysics physics;
  final EdgeInsets padding;
  final bool dismissKeyBoard;

  Scroll2({
    Scroll2Controller bus,
    @required this.child,
    this.physics = const BouncingScrollPhysics(parent: AlwaysScrollableScrollPhysics()),
    this.padding = EdgeInsets.zero,
    this.dismissKeyBoard = false,
    this.axis = Axis.horizontal,
  }) : controller = bus ?? Scroll2Controller();

  addScrollListener({
    @required String tag,
    @required double max,
    double min = 0,
    Function(double offset) scroll,
    Function(double offset) forwardScroll,
    Function(double offset) reverseScroll,
  }) {
    return controller.subscribeToScroll(
      tag: tag,
      min: min,
      max: max,
      scroll: scroll,
      forwardScroll: forwardScroll,
      reverseScroll: reverseScroll,
    );
  }

  addOverscrollListener({
    @required String tag,
    Function(double overscroll) startOverscroll,
    Function(double overscroll) endOverscroll,
  }) {
    return controller.subscribeToOverscroll(
      tag: tag,
      endOverscroll: endOverscroll,
      startOverscroll: startOverscroll,
    );
  }

  addScrollStateListener({
    @required String tag,
    @required double max,
    double min = 0,
    Function(double offset) scroll,
    Function(double offset) forwardScroll,
    Function(double offset) reverseScroll,
  }) {
    return controller.subscribeToScroll(
      tag: tag,
      min: min,
      max: max,
      scroll: scroll,
      forwardScroll: forwardScroll,
      reverseScroll: reverseScroll,
    );
  }

  @override
  Widget get body {
    return SingleChildScrollView(
      child: child ?? empty,
      physics: physics ?? const BouncingScrollPhysics(parent: AlwaysScrollableScrollPhysics()),
      padding: padding ?? EdgeInsets.zero,
      clipBehavior: Clip.none,
      scrollDirection: axis,
      controller: controller,
    );
  }

  @override
  StructState<Struct> createState() => _Scroll2State();
}

class _Scroll2State<T extends Scroll2> extends StructState<T> {
  @override
  void initState() {
    widget.controller.activate();
    super.initState();
  }

  @override
  void dispose() {
    widget.controller.deactivate();
    super.dispose();
  }
}

/// <h1>>> ScrollBus </h1>
/// A more advance [ScrollController] to handle scroll and overscroll events in a [Scroll2]. <br>
class Scroll2Controller extends ScrollController {
  bool _activated = false;
  int reps = 0;

  ScrollDirection _formerDirection = ScrollDirection.idle;
  double _refScrollOffset = 0;
  double _refScrollChange = 0;
  double _formerOffset = 0;

  Map<String, Function()> _idle = {};
  Map<String, Function()> _active = {};
  Map<String, Function()> _forwardScrollStart = {};
  Map<String, Function()> _reverseScrollStart = {};
  Map<String, Function(double offset)> _scroll = {};
  Map<String, Function(double offset)> _forwardScroll = {};
  Map<String, Function(double offset)> _reverseScroll = {};
  Map<String, double> _max = {};
  Map<String, double> _min = {};

  /// offset is the positive value from the restricted boundary
  Map<String, Function(double overscroll)> _startOverscroll = {};

  /// offset is the positive value from the restricted boundary
  Map<String, Function(double overscroll)> _endOverscroll = {};

  Scroll2Controller();

  String subscribeToScroll({
    @required String tag,
    @required double max,
    double min = 0,
    Function(double offset) scroll,
    Function(double offset) forwardScroll,
    Function(double offset) reverseScroll,
  }) {
    this._max.update(tag + 'stc', (value) => max, ifAbsent: () => max);
    this._min.update(tag + 'stc', (value) => min, ifAbsent: () => min);
    if (scroll != null) _scroll.update(tag, (value) => scroll, ifAbsent: () => scroll);
    if (forwardScroll != null) _forwardScroll.update(tag, (value) => forwardScroll, ifAbsent: () => forwardScroll);
    if (reverseScroll != null) _reverseScroll.update(tag, (value) => reverseScroll, ifAbsent: () => reverseScroll);
    return tag;
  }

  String subscribeToOverscroll({
    @required String tag,
    Function(double overscroll) startOverscroll,
    Function(double overscroll) endOverscroll,
  }) {
    if (startOverscroll != null) _startOverscroll.update(tag, (value) => startOverscroll, ifAbsent: () => startOverscroll);
    if (endOverscroll != null) _endOverscroll.update(tag, (value) => endOverscroll, ifAbsent: () => endOverscroll);
    return tag;
  }

  String subscribeToDirection({
    @required String tag,
    Function() idle,
    Function() active,
    Function() forward,
    Function() reverse,
  }) {
    if (idle != null) _idle.update(tag, (value) => idle, ifAbsent: () => idle);
    if (active != null) _active.update(tag, (value) => active, ifAbsent: () => active);
    if (forward != null) _forwardScrollStart.update(tag, (value) => forward, ifAbsent: () => forward);
    if (reverse != null) _reverseScrollStart.update(tag, (value) => reverse, ifAbsent: () => reverse);
    return tag;
  }

  /// This method is still under development
  @deprecated
  String subscribeToRange({
    @required String tag,
    double min = 0,
    @required double max,
    Function(bool) callback,
  }) {
    //TODO Implement
    throw UnimplementedError('This method is still under development');
  }

  Function() get compile {
    return () {
      ScrollDirection dir = position.userScrollDirection;
      if (dir == ScrollDirection.forward) {
        if (_formerDirection != ScrollDirection.forward) {
          _refScrollOffset = offset;
          _refScrollChange = _formerOffset;
          _forwardScrollStart.forEach((key, value) => value());
          if (_formerDirection == ScrollDirection.idle) _active.forEach((key, value) => value());
        }
        if (offset < 0) {
          if (offset != _formerOffset) {
            _startOverscroll.forEach((key, value) => value(-offset));
          }
        } else if (offset > position.maxScrollExtent) {
          if (offset != _formerOffset) {
            _scroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
            _endOverscroll.forEach((key, value) => value(offset - position.maxScrollExtent));
          }
        } else {
          if (offset != _formerOffset) {
            _scroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
            _forwardScroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
          }
        }
      } else if (dir == ScrollDirection.reverse) {
        if (_formerDirection != ScrollDirection.reverse) {
          _refScrollOffset = offset;
          _refScrollChange = _formerOffset;
          _reverseScrollStart.forEach((key, value) => value());
          if (_formerDirection == ScrollDirection.idle) _active.forEach((key, value) => value());
        }
        if (offset < 0) {
          if (offset != _formerOffset) {
            _startOverscroll.forEach((key, value) => value(-offset));
          }
        } else if (offset > position.maxScrollExtent) {
          if (offset != _formerOffset) {
            _scroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
            _endOverscroll.forEach((key, value) => value(offset - position.maxScrollExtent));
          }
        } else {
          if (offset != _formerOffset) {
            _scroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
            _reverseScroll.forEach((key, value) => value(min(_max[key + 'stc'], max(0, _refScrollChange + (offset - _refScrollOffset)))));
          }
        }
      } else if (dir == ScrollDirection.idle) {
        if (_formerDirection != ScrollDirection.idle) _idle.forEach((key, value) => value());
      }
      _formerDirection = dir;
      _formerOffset = offset;
    };
  }

  unsubscribe(String tag) {
    _max.removeWhere((key, value) => key == tag + 'stc');
    _min.removeWhere((key, value) => key == tag + 'stc');
    _max.removeWhere((key, value) => key == tag + 'sto');
    _min.removeWhere((key, value) => key == tag + 'sto');
    _idle.removeWhere((key, value) => key == tag);
    _active.removeWhere((key, value) => key == tag);
    _scroll.removeWhere((key, value) => key == tag);
    _forwardScroll.removeWhere((key, value) => key == tag);
    _reverseScroll.removeWhere((key, value) => key == tag);
    _forwardScrollStart.removeWhere((key, value) => key == tag);
    _reverseScrollStart.removeWhere((key, value) => key == tag);
    _endOverscroll.removeWhere((key, value) => key == tag);
    _startOverscroll.removeWhere((key, value) => key == tag);
    deactivate(thorough: true);
    activate(enforce: true);
  }

  activate({bool enforce = false}) {
    if (enforce || !_activated) {
      _activated = true;
      addListener(compile);
      reps++;
    }
  }

  bool get isActivated => _activated;

  deactivate({bool thorough = false}) {
    int _reps = reps - 1;
    if (_activated) {
      while (reps > (thorough ? 0 : _reps)) {
        reps--;
        removeListener(compile);
      }
      _activated = false;
    }
  }
}
